#include "include.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>


/* RTcmix defines below -- can't include all the RTcmix .h
   files because of symbol-naming conflicts */

#define MAXDISPARGS 1024 /* from cmix/Minc/defs.h */
#define MAXTEXTARGS ((MAXDISPARGS*sizeof(double))/512)

struct sockdata {
	char name[100];		/* the cmix command name */
	union {
		double p[MAXDISPARGS];	/* the p-fields */
		char text[MAXTEXTARGS][512];     /* or text arrays  */
		} data;
	int n_args;		/* number of p-fields used */
	};



/*
*  (rtcmix-socket-open "hostname" sockno) 
*	-- sockno is TCP/IP socket numebr, usually 1102 for RTcmix
*	   instruments (the default)
*  returns the socket filedesc for use in rtcmix-print(), etc.
*/
DEFUN("RTCMIX-SOCKET-OPEN",int,fLrtcmix_socket_open,LISP,2,2,NONE,
	IO,IO,OO,OO,"open a socket to RTcmix")
	(hname,sockno)
object hname;
int sockno;
{
	int s;
	struct sockaddr_in sss;
	struct hostent *hp;

        if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		printf("problem setting the socket\n");
		return (-1);
		}

	hp = gethostbyname(hname->st.st_self);
	sss.sin_family = AF_INET;
	bcopy(hp->h_addr, &(sss.sin_addr.s_addr), hp->h_length);
	sss.sin_port = htons(sockno);

	if(connect(s, (struct sockaddr *)&sss, sizeof(sss)) < 0) {
		printf("problem connecting to socket: %d\n",sockno);
		printf("   on host: %s\n",hname->st.st_self);
		return (-1);
		}

	return s;
}

/*
*  (rtcmix-socket-setblock sockno)
*	-- sets sockno to block on reading or writing
*/
DEFUN("RTCMIX-SOCKET-SETBLOCK",object,fLrtcmix_socket_setblock,
	LISP,1,1,NONE,OO,OO,OO,OO,"set socket to block")
	(sockno)
object sockno;
{
	if (fcntl(fix(sockno), F_SETFL, 0) == -1) {
		printf("error: rtcmix-socket-setblock -- fcntl to blocking mode\n");
		return Cnil;
		}

	return Ct;
}

/*
*  (rtcmix-socket-setnblock sockno)
*	-- sets sockno to block on reading or writing
*/
DEFUN("RTCMIX-SOCKET-SETNBLOCK",object,fLrtcmix_socket_setnblock,
	LISP,1,1,NONE,OO,OO,OO,OO,"set socket to nonblock")
	(sockno)
object sockno;
{
	if (fcntl(fix(sockno), F_SETFL, FNONBLK) == -1) {
		printf("error: rtcmix-socket-setnblock -- fcntl to non-blocking mode\n");
		return Cnil;
		}

	return Ct;
}


/*
*  (rtcmix-print "command" (params) RTcmixsocket)
*	-- (params) is a list of params to be set as p-fields
*	   RTcmixsocket is the socket returned by rtcmix-socket-open
*  returns T or NIL depending on write success status
*/
DEFUN("RTCMIX-PRINT",object,fLrtcmix_print,LISP,3,3,NONE,
	OO,OO,OO,OO,"write to an RTcmix instrument")
	(command,params,sock)
object command,params,sock;
{
	struct sockdata ssend;
	object tparams,tp;
	int i;

	strcpy(ssend.name, command->st.st_self);
/* for some reason lisp does weird things if command->st.st_self < 5 chars */
	if (strncmp(ssend.name, "load", 4) == 0) strcpy(ssend.name, "load");

	i = 0;
	for (tparams = params; !endp(tparams); tparams = MMcdr(tparams)) {
		tp = MMcar(tparams);
		switch (type_of(tp)) {
			case t_fixnum:
				ssend.data.p[i] = fix(tp);
				break;
			case t_bignum:
				ssend.data.p[i] = fix(tp);
				break;
			case t_shortfloat:
				ssend.data.p[i] = sf(tp);
				break;
			case t_longfloat:
				ssend.data.p[i] = lf(tp);
				break;
			case t_cons:
				printf("error: RTcmix -- nested list in params\n");
				return Cnil;
			case t_string:
				strcpy(ssend.data.text[i], tp->st.st_self);
				break;
			}
		i++;
		}
	ssend.n_args = i;

	if (write(fix(sock), (char *)&ssend, sizeof(struct sockdata)) < 0) {
		printf("error: Rtcmix -- writing to socket\n");
		return Cnil;
		}

	return Ct;
}



int sleeping;

/*
*  (rtcmix-timer timeval)
*	-- timeval is the (can be floating point!) number of
*	seconds to suspend execution
*  returns T
*/
DEFUN("RTCMIX-TIMER",object,fLrtcmix_timer,LISP,1,1,NONE,
	OO,OO,OO,OO,"sleep for secs + microsecs time")
	(tval)
object tval;
{
	int nsecs,nusecs;
	void awake();
	struct timeval tv;
	struct itimerval itv;

	switch (type_of(tval)) {
		case t_fixnum:
			nsecs = fix(tval);
			nusecs = 0;
			break;
		case t_bignum:
			nsecs = fix(tval);
			nusecs = 0;
			break;
		case t_shortfloat:
			nsecs = (int)sf(tval);
			nusecs = (int)((lf(tval) - (float)nsecs) * 1000000.0);
			break;
                case t_longfloat:
			nsecs = (int)lf(tval);
			nusecs = (int)((lf(tval) - (float)nsecs) * 1000000.0);
			break;
		case t_cons:
			printf("rtcmix-timer: illegal list\n");
			return Cnil;
                case t_string:
			printf("rtcmix-timer: illegal string\n");
			return Cnil;
		}

	if ((nsecs == 0) && (nusecs == 0)) return Ct;

	tv.tv_sec = nsecs;
	tv.tv_usec = nusecs;
	itv.it_interval = tv;
	itv.it_value = tv;

	setitimer(ITIMER_REAL, &itv, NULL);
	sigset(SIGALRM, awake);

	sleeping = 1; /* reset by awake() when timer expires */
	while(sleeping == 1) sleep(1000);

	return Ct;
}

void
awake()
{
	struct timeval tv;
	struct itimerval itv;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	itv.it_interval = tv;
	itv.it_value = tv;
	setitimer(ITIMER_REAL, &itv, NULL);

	sleeping = 0;
}




/* The following are functions designed to use an external interface:
*  values are stored in the 2-dimensional array ifvals[][],
*  updated through rtcmix-setifvals(socket), and retrieved via a call
*  to rtcmix-getifval(ifno).
*/

int ifvals[100][1];

/*
*  (rtcmix-getifval val) -- returns the int the value from ifvals[val][0]
*/
DEFUN("RTCMIX-GETIFVAL",int,fLrtcmix_getifval,LISP,1,1,NONE,
	II,OO,OO,OO,"retrieve an ifval from the ifvals array")
	(kval)
int kval;
{
	return ifvals[kval][0];
}


/*
*   (rtcmix-setifvals sockno) -- reads incoming values from socket
*	sockno into the ifvals[][] array
*   returns T if any updated, NIL if not
*/
DEFUN("RTCMIX-SETIFVALS",object,fLrtcmix_setifvals,LISP,1,1,NONE,
	OO,OO,OO,OO,"read ifvals from a socket")
	(sockno)
object sockno;
{
	int s,ivals[2];
	int incoming,fval;
	int i,setvals[100],nset,sflag;
	object tobj;

	s = fix(sockno);

	fval = fcntl(s, F_GETFL);
	incoming = 0;
	nset = 0;
	
	if (fval & FNONBLK) {
		while (read(s, (void *)ivals, sizeof(ivals)) > 0) {
			ifvals[ivals[0]][0] = ivals[1];

			sflag = 1;
			for (i = 0; i < nset; i++) {
				if (ivals[0] == setvals[i]) sflag = 0;
				}
			if (sflag) setvals[nset++] = ivals[0];
			incoming = 1;
			}
		}
	else {
		read(s, (void *)ivals, sizeof(ivals));
		ifvals[ivals[0]][0] = ivals[1];
		setvals[nset++] = ivals[0];
		incoming = 1;
		}

	tobj = Cnil;
	if (incoming) {
		for (i = 0; i < nset; i++) 
			tobj = make_cons(make_fixnum(setvals[i]), tobj);
			}
	return tobj;
}


